from dkg.constants import (
    DefaultParameters,
    ZERO_ADDRESS,
    DEFAULT_PROXIMITY_SCORE_FUNCTIONS_PAIR_IDS,
)


class InputService:
    def __init__(self, manager, config):
        self.manager = manager
        self.config = config

    def get_asset_get_arguments(self, options):
        return {
            "max_number_of_retries": self.get_max_number_of_retries(options),
            "frequency": self.get_frequency(options),
            "state": self.get_state(options),
            "include_metadata": self.get_include_metadata(options),
            "content_type": self.get_content_type(options),
            "validate": self.get_validate(options),
            "output_format": self.get_output_format(options),
            "hash_function_id": self.get_hash_function_id(options),
            "paranet_ual": self.get_paranet_ual(options),
            "subject_ual": self.get_subject_ual(options),
        }

    def get_asset_create_arguments(self, options):
        return {
            "max_number_of_retries": self.get_max_number_of_retries(options),
            "frequency": self.get_frequency(options),
            "epochs_num": self.get_epochs_num(options),
            "hash_function_id": self.get_hash_function_id(options),
            "score_function_id": self.get_score_function_id(options),
            "immutable": self.get_immutable(options),
            "token_amount": self.get_token_amount(options),
            "payer": self.get_payer(options),
            "minimum_number_of_finalization_confirmations": self.get_minimum_number_of_finalization_confirmations(
                options
            ),
            "minimum_number_of_node_replications": self.get_minimum_number_of_node_replications(
                options
            ),
        }

    def get_query_arguments(self, options):
        return {
            "paranet_ual": self.get_paranet_ual(options),
            "repository": self.get_repository(options),
        }

    def get_publish_finality_arguments(self, options):
        return {
            "max_number_of_retries": self.get_max_number_of_retries(options),
            "frequency": self.get_frequency(options),
            "minimum_number_of_finalization_confirmations": self.get_minimum_number_of_finalization_confirmations(
                options
            ),
        }

    def get_max_number_of_retries(self, options):
        return (
            options.get("max_number_of_retries")
            or self.config.get("max_number_of_retries")
            or DefaultParameters.MAX_NUMBER_OF_RETRIES.value
        )

    def get_frequency(self, options):
        return (
            options.get("frequency")
            or self.config.get("frequency")
            or DefaultParameters.FREQUENCY.value
        )

    def get_state(self, options):
        return (
            options.get("state")
            or self.config.get("state")
            or DefaultParameters.STATE.value
        )

    def get_include_metadata(self, options):
        return (
            options.get("include_metadata")
            or self.config.get("include_metadata")
            or DefaultParameters.INCLUDE_METADATA.value
        )

    def get_content_type(self, options):
        return (
            options.get("content_type")
            or self.config.get("content_type")
            or DefaultParameters.CONTENT_TYPE.value
        )

    def get_validate(self, options):
        return (
            options.get("validate")
            or self.config.get("validate")
            or DefaultParameters.VALIDATE.value
        )

    def get_output_format(self, options):
        return (
            options.get("output_format")
            or self.config.get("output_format")
            or DefaultParameters.OUTPUT_FORMAT.value
        )

    def get_hash_function_id(self, options):
        return (
            options.get("hash_function_id")
            or self.config.get("hash_function_id")
            or DefaultParameters.HASH_FUNCTION_ID.value
        )

    def get_paranet_ual(self, options):
        return (
            options.get("paranet_ual")
            or self.config.get("paranet_ual")
            or DefaultParameters.PARANET_UAL.value
        )

    def get_subject_ual(self, options):
        return (
            options.get("subject_ual")
            or self.config.get("subject_ual")
            or DefaultParameters.GET_SUBJECT_UAL.value
        )

    def get_epochs_num(self, options):
        return options.get("epochs_num") or self.config.get("epochs_num") or None

    def get_immutable(self, options):
        return (
            options.get("immutable")
            or self.config.get("immutable")
            or DefaultParameters.IMMUTABLE.value
        )

    def get_token_amount(self, options):
        return options.get("token_amount") or self.config.get("token_amount") or None

    def get_payer(self, options):
        return options.get("payer") or self.config.get("payer") or ZERO_ADDRESS

    def get_minimum_number_of_finalization_confirmations(self, options):
        return (
            options.get("minimum_number_of_finalization_confirmations")
            or self.config.get("minimum_number_of_finalization_confirmations")
            or DefaultParameters.MIN_NUMBER_OF_FINALIZATION_CONFIRMATION.value
            or None
        )

    def get_minimum_number_of_node_replications(self, options):
        return (
            options.get("minimum_number_of_node_replications")
            or self.config.get("minimum_number_of_node_replications")
            or None
        )

    def get_score_function_id(self, options):
        environment = (
            options.get("environment")
            or self.config.get("environment")
            or self.manager.blockchain_provider.environment
            or DefaultParameters.ENVIRONMENT.value
        )
        blockchain_name = (
            options.get("blockchain")
            or self.config.get("blockchain")
            or self.manager.blockchain_provider.blockchain_id
        )

        return DEFAULT_PROXIMITY_SCORE_FUNCTIONS_PAIR_IDS[environment][blockchain_name]

    def get_repository(self, options):
        return (
            options.get("repository")
            or self.config.get("repository")
            or DefaultParameters.REPOSITORY.value
        )



from dkg.managers.manager import DefaultRequestManager
from dkg.method import Method
from dkg.modules.module import Module
import time
from dkg.utils.decorators import retry
from dkg.exceptions import (
    OperationNotFinished,
)
from dkg.utils.node_request import (
    NodeRequest,
    validate_operation_status,
)


class NodeService(Module):
    def __init__(self, manager: DefaultRequestManager):
        self.manager = manager

    _get_operation_result = Method(NodeRequest.get_operation_result)
    _finality_status = Method(NodeRequest.finality_status)
    _ask = Method(NodeRequest.ask)
    _publish = Method(NodeRequest.publish)
    _get = Method(NodeRequest.get)
    _query = Method(NodeRequest.query)

    def get_operation_result(
        self, operation_id: str, operation: str, max_retries: int, frequency: int
    ):
        @retry(
            catch=OperationNotFinished,
            max_retries=max_retries,
            base_delay=frequency,
            backoff=1,
        )
        def retry_get_operation_result():
            operation_result = self._get_operation_result(
                operation_id=operation_id,
                operation=operation,
            )
            validate_operation_status(operation_result)

            return operation_result

        return retry_get_operation_result()

    def finality_status(
        self,
        ual: str,
        required_confirmations: int,
        max_number_of_retries: int,
        frequency: int,
    ):
        retries = 0
        finality = 0

        while finality < required_confirmations and retries <= max_number_of_retries:
            if retries > max_number_of_retries:
                raise Exception(
                    f"Unable to achieve required confirmations. "
                    f"Max number of retries ({max_number_of_retries}) reached."
                )

            if retries > 0:
                time.sleep(frequency)

            retries += 1

            try:
                try:
                    response = self._finality_status(ual=ual)
                except Exception as e:
                    response = None
                    print(f"failed: {e}")

                if response is not None:
                    finality = response.get("finality", 0)
                    if finality >= required_confirmations:
                        break

            except Exception:
                finality = 0

        return finality

    def ask(self, ual, required_confirmations, max_number_of_retries, frequency):
        confirmations_count = 0
        retries = 0

        while (
            confirmations_count < required_confirmations
            and retries < max_number_of_retries
        ):
            if retries > max_number_of_retries:
                raise Exception(
                    f"Unable to achieve required confirmations. "
                    f"Max number of retries ({max_number_of_retries}) reached."
                )

            if retries > 0:
                time.sleep(frequency)

            retries += 1

            try:
                try:
                    response = self._ask(
                        ual=ual, minimumNumberOfNodeReplications=required_confirmations
                    )
                except Exception as e:
                    response = None
                    print(f"failed: {e}")

                if response is not None:
                    number_of_confirmations = response.json().get(
                        "numberOfConfirmations", 0
                    )
                    if number_of_confirmations >= required_confirmations:
                        confirmations_count = number_of_confirmations

            except Exception as e:
                confirmations_count = 0
                print(f"Retry {retries + 1}/{max_number_of_retries} failed: {e}")

            return confirmations_count

    def publish(
        self,
        dataset_root,
        dataset,
        blockchain_id,
        hash_function_id,
        minimum_number_of_node_replications,
    ):
        return self._publish(
            dataset_root,
            dataset,
            blockchain_id,
            hash_function_id,
            minimum_number_of_node_replications,
        )

    def get(
        self,
        ual_with_state,
        content_type,
        include_metadata,
        hash_function_id,
        paranet_ual,
        subject_ual,
    ):
        return self._get(
            ual_with_state,
            content_type,
            include_metadata,
            hash_function_id,
            paranet_ual,
            subject_ual,
        )

    def query(
        self,
        query,
        query_type,
        repository,
        paranet_ual,
    ):
        return self._query(query, query_type, repository, paranet_ual)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from rdflib.plugins.sparql.parser import parseQuery

from dkg.managers.async_manager import AsyncRequestManager
from dkg.modules.async_module import AsyncModule
from dkg.types import NQuads
from dkg.constants import Status
from dkg.services.input_service import InputService
from dkg.services.node_services.async_node_service import AsyncNodeService
from dkg.types import UAL


class AsyncGraph(AsyncModule):
    def __init__(
        self,
        manager: AsyncRequestManager,
        input_service: InputService,
        node_service: AsyncNodeService,
    ):
        self.manager = manager
        self.input_service = input_service
        self.node_service = node_service

    async def query(
        self,
        query: str,
        options: dict = None,
    ) -> NQuads:
        if options is None:
            options = {}

        arguments = self.input_service.get_query_arguments(options)

        paranet_ual = arguments.get("paranet_ual")
        repository = arguments.get("repository")

        parsed_query = parseQuery(query)
        query_type = parsed_query[1].name.replace("Query", "").upper()

        result = await self.node_service.query(
            query, query_type, repository, paranet_ual
        )
        result.encoding = "utf-8"

        return result.get("data")

    async def publish_finality(self, UAL: UAL, options=None):
        if options is None:
            options = {}

        arguments = self.input_service.get_publish_finality_arguments(options)
        max_number_of_retries = arguments.get("max_number_of_retries")
        minimum_number_of_finalization_confirmations = arguments.get(
            "minimum_number_of_finalization_confirmations"
        )
        frequency = arguments.get("frequency")
        try:
            finality_status_result = await self.node_service.finality_status(
                UAL,
                minimum_number_of_finalization_confirmations,
                max_number_of_retries,
                frequency,
            )
        except Exception as e:
            return {"status": Status.ERROR.value, "error": str(e)}

        if finality_status_result == 0:
            try:
                finality_operation_id = await self.node_service.finality(
                    UAL,
                    minimum_number_of_finalization_confirmations,
                    max_number_of_retries,
                    frequency,
                )
            except Exception as e:
                return {"status": Status.ERROR.value, "error": str(e)}

            try:
                return await self.node_service.get_operation_result(
                    finality_operation_id, "finality", max_number_of_retries, frequency
                )
            except Exception as e:
                return {"status": Status.NOT_FINALIZED.value, "error": str(e)}

        elif finality_status_result >= minimum_number_of_finalization_confirmations:
            return {
                "status": Status.FINALIZED.value,
                "numberOfConfirmations": finality_status_result,
                "requiredConfirmations": minimum_number_of_finalization_confirmations,
            }
        else:
            return {
                "status": Status.NOT_FINALIZED.value,
                "numberOfConfirmations": finality_status_result,
                "requiredConfirmations": minimum_number_of_finalization_confirmations,
            }



from typing import Literal
import codecs
from dkg.constants import CHUNK_BYTE_SIZE
from dkg.exceptions import DatasetInputFormatNotSupported, InvalidDataset
from dkg.types import JSONLD, NQuads
from pyld import jsonld
from dkg.constants import DEFAULT_RDF_FORMAT, DEFAULT_CANON_ALGORITHM, ESCAPE_MAP
from rdflib import Graph, BNode, URIRef, Dataset
from rdflib.exceptions import ParserError as RDFParserError
from uuid import uuid4
from web3 import Web3
import math
import hashlib
from eth_abi.packed import encode_packed


def normalize_dataset(
    dataset: JSONLD | NQuads,
    input_format: Literal["JSON-LD", "N-Quads"] = "JSON-LD",
    output_format=DEFAULT_RDF_FORMAT,
    algorithm=DEFAULT_CANON_ALGORITHM,
) -> NQuads:
    normalization_options = {
        "algorithm": algorithm,
        "format": output_format,
    }

    match input_format.lower():
        case "json-ld" | "jsonld":
            pass
        case "n-quads" | "nquads":
            normalization_options["inputFormat"] = "application/n-quads"
        case _:
            raise DatasetInputFormatNotSupported(
                f"Dataset input format isn't supported: {input_format}. "
                "Supported formats: JSON-LD / N-Quads."
            )

    n_quads = jsonld.normalize(dataset, normalization_options)
    assertion = [quad for quad in n_quads.split("\n") if quad]

    if not assertion:
        raise InvalidDataset("Invalid dataset, no quads were extracted.")

    return assertion


def is_empty_dict(dictionary: dict):
    return len(dictionary.keys()) == 0 and isinstance(dictionary, dict)


def format_dataset(
    content: dict,
    input_format: Literal["JSON-LD", "N-Quads"] = "JSON-LD",
    output_format=DEFAULT_RDF_FORMAT,
    algorithm=DEFAULT_CANON_ALGORITHM,
):
    private_assertion = None
    if content.get("private") and not is_empty_dict(content.get("private")):
        private_assertion = normalize_dataset(
            content.get("private"), input_format, output_format, algorithm
        )
    elif not content.get("public"):
        content = {"public": content}

    public_assertion = []
    if content.get("public"):
        public_assertion = normalize_dataset(
            content.get("public"), input_format, output_format, algorithm
        )

    if (
        public_assertion
        and len(public_assertion) == 0
        and private_assertion
        and len(private_assertion) == 0
    ):
        raise ValueError("File format is corrupted, no n-quads are extracted.")

    dataset = {"public": public_assertion}
    if private_assertion:
        dataset["private"] = private_assertion

    return dataset


def split_into_chunks(quads, chunk_size_bytes=32):
    # Concatenate the quads with newline characters
    concatenated_quads = "\n".join(quads)

    # Encode the concatenated string to bytes
    encoded_bytes = concatenated_quads.encode("utf-8")

    # Split the encoded bytes into chunks
    chunks = []
    start = 0

    while start < len(encoded_bytes):
        end = min(start + chunk_size_bytes, len(encoded_bytes))
        chunk = encoded_bytes[start:end]
        chunks.append(
            codecs.decode(chunk, "utf-8", errors="replace")
        )  # Decode bytes back to string
        start = end

    return chunks


def calculate_merkle_root(quads: list[str], chunk_size_bytes: int = CHUNK_BYTE_SIZE):
    chunks = split_into_chunks(quads, chunk_size_bytes)

    # Create leaves using solidityKeccak256 equivalent
    leaves = [
        bytes.fromhex(Web3.solidity_keccak(["string", "uint256"], [chunk, index]).hex())
        for index, chunk in enumerate(chunks)
    ]

    while len(leaves) > 1:
        next_level = []

        for i in range(0, len(leaves), 2):
            left = leaves[i]

            if i + 1 >= len(leaves):
                next_level.append(left)
                break

            right = leaves[i + 1]

            # Combine and sort the leaves
            combined = [left, right]
            combined.sort()

            # Calculate the hash of the combined leaves
            hash_value = Web3.keccak(b"".join(combined))
            next_level.append(hash_value)

        leaves = next_level

    return f"0x{leaves[0].hex()}"


def generate_missing_ids_for_blank_nodes(nquads_list: list[str] | None) -> list[str]:
    if not nquads_list:
        return [""]

    generated_ids = {}

    def replace_blank_node(term):
        # Handle blank nodes
        if isinstance(term, BNode):
            if str(term) not in generated_ids:
                generated_ids[str(term)] = URIRef(f"uuid:{str(uuid4())}")
            return generated_ids[str(term)]

        return term  # Return IRIs or Literals unchanged

    all_nquads = "\n".join(nquad for nquad in nquads_list if nquad.strip())

    # Create a single Dataset
    g = Graph()
    try:
        g.parse(data=all_nquads, format="nt")
    except RDFParserError:
        raise UnsupportedJSONLD(nquads_list)

    # Process all quads
    result = []
    for s, p, o in g:
        updated_quad = (
            replace_blank_node(s),
            replace_blank_node(p),
            replace_blank_node(o),
        )
        result.append(
            f"{updated_quad[0].n3()} {updated_quad[1].n3()} {updated_quad[2].n3()} ."
        )

    return result


def group_nquads_by_subject(nquads_list: list[str], sort: bool = False):
    grouped = {}

    all_nquads = "\n".join(nquad for nquad in nquads_list if nquad.strip())

    d = Dataset()
    d.parse(data=all_nquads, format="nquads")

    for quad in d:
        subject, predicate, obj, graph = quad

        # Get subject key
        subject_key = (
            f"<<<{subject.subject}> <{subject.predicate}> <{subject.object}>>"
            if hasattr(subject, "subject")
            else f"<{subject}>"
        )

        # Initialize group if needed
        if subject_key not in grouped:
            grouped[subject_key] = []

        # Add quad to group
        quad_string = f"{subject.n3()} {predicate.n3()} {obj.n3()} ."
        grouped[subject_key].append(quad_string)

    # Return grouped quads (sorted if requested)
    grouped_items = sorted(grouped.items()) if sort else grouped.items()
    return [quads for _, quads in grouped_items]


def calculate_number_of_chunks(quads, chunk_size_bytes=CHUNK_BYTE_SIZE):
    # Concatenate the quads with newline characters
    concatenated_quads = "\n".join(quads)

    total_size_bytes = len(concatenated_quads.encode("utf-8"))

    # Calculate and return the number of chunks
    return math.ceil(total_size_bytes / chunk_size_bytes)


def count_distinct_subjects(nquads_list: list[str]) -> int:
    # Create a new RDF graph
    graph = Graph()

    # Parse the joined N-Quads
    graph.parse(data="\n".join(nquads_list), format="nquads")

    # Extract unique subjects using set comprehension
    subjects = {str(quad[0]) for quad in graph}

    return len(subjects)


def solidity_packed_sha256(types: list[str], values: list) -> str:
    # Encode the values using eth_abi's encode_packed
    packed_data = encode_packed(types, values)

    # Calculate SHA256
    sha256_hash = hashlib.sha256(packed_data).hexdigest()

    return f"0x{sha256_hash}"


def escape_literal_string(s):
    for char, replacement in ESCAPE_MAP.items():
        s = s.replace(char, replacement)
    return s


def escape_literal_dict(obj):
    if isinstance(obj, dict):
        return {k: escape_literal_dict(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [escape_literal_dict(i) for i in obj]
    elif isinstance(obj, str):
        return escape_literal_string(s=obj)
    else:
        return obj


# Used when JSON-LD parsing fails due to quads being passed instead of triples
class UnsupportedJSONLD(Exception):
    def __init__(self, nquads_list):
        self.nquads_list = nquads_list
        self.message = f"""
Unsupported JSON-LD input detected

After parsing the JSON-LD input, the parser detected creation of new named graphs.
The DKG does not support custom named graphs.

Problematic Quads:

{self.find_problematic_quads()}

Full Parsed N-Quads Array:

{self.format_nquads_list()}

"""
        super().__init__(self.message)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"

    def format_nquads_list(self):
        return "\n".join(nquad.strip() for nquad in self.nquads_list)

    def find_problematic_quads(self):
        problematic = []
        g = Graph()
        for quad in self.nquads_list:
            if not quad.strip():
                continue
            try:
                g.parse(data=quad, format="nt")
            except RDFParserError:
                problematic.append(quad)

        return "\n".join(f"{i + 1}. {quad}" for i, quad in enumerate(problematic))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json

from dataclasses import dataclass
from web3 import Web3
from web3.types import TxReceipt

from dkg.dataclasses import (
    BaseIncentivesPoolParams,
    ParanetIncentivizationType,
    ParanetNodesAccessPolicy,
    ParanetMinersAccessPolicy,
)
from dkg.managers.manager import DefaultRequestManager
from dkg.method import Method
from dkg.modules.module import Module
from dkg.types import Address, UAL, HexStr
from dkg.utils.blockchain_request import BlockchainRequest
from dkg.utils.ual import parse_ual


class Paranet(Module):
    @dataclass
    class NeuroWebIncentivesPoolParams(BaseIncentivesPoolParams):
        neuro_emission_multiplier: float
        operator_percentage: float
        voters_percentage: float

        def to_contract_args(self) -> dict:
            return {
                "tracToNeuroEmissionMultiplier": int(
                    self.neuro_emission_multiplier * (10**12)
                ),
                "paranetOperatorRewardPercentage": int(self.operator_percentage * 100),
                "paranetIncentivizationProposalVotersRewardPercentage": int(
                    self.voters_percentage * 100
                ),
            }

    def __init__(self, manager: DefaultRequestManager):
        self.manager = manager
        self.incentives_pools_deployment_functions = {
            ParanetIncentivizationType.NEUROWEB: self._deploy_neuro_incentives_pool,
        }

    _register_paranet = Method(BlockchainRequest.register_paranet)

    def create(
        self,
        ual: UAL,
        name: str,
        description: str,
        paranet_nodes_access_policy: ParanetNodesAccessPolicy,
        paranet_miners_access_policy: ParanetMinersAccessPolicy,
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt: TxReceipt = self._register_paranet(
            knowledge_asset_storage,
            knowledge_asset_token_id,
            name,
            description,
            paranet_nodes_access_policy,
            paranet_miners_access_policy,
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _add_paranet_curated_nodes = Method(BlockchainRequest.add_paranet_curated_nodes)

    def add_curated_nodes(
        self, paranet_ual: UAL, identity_ids: list[int]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._add_paranet_curated_nodes(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            identity_ids,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _remove_paranet_curated_nodes = Method(
        BlockchainRequest.remove_paranet_curated_nodes
    )

    def remove_curated_nodes(
        self, paranet_ual: UAL, identity_ids: list[int]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._remove_paranet_curated_nodes(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            identity_ids,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _request_paranet_curated_node_access = Method(
        BlockchainRequest.request_paranet_curated_node_access
    )

    def request_curated_node_access(
        self, paranet_ual: UAL
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._request_paranet_curated_node_access(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _approve_curated_node = Method(BlockchainRequest.approve_curated_node)

    def approve_curated_node(
        self, paranet_ual: UAL, identity_id: int
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._approve_curated_node(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            identity_id,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _reject_curated_node = Method(BlockchainRequest.reject_curated_node)

    def reject_curated_node(
        self, paranet_ual: UAL, identity_id: int
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._reject_curated_node(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            identity_id,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _get_curated_nodes = Method(BlockchainRequest.get_curated_nodes)

    def get_curated_nodes(
        self, paranet_ual: UAL
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        paranet_id = Web3.solidity_keccak(
            ["address", "uint256"],
            [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
        )

        return self._get_curated_nodes(paranet_id)

    _add_paranet_curated_miners = Method(BlockchainRequest.add_paranet_curated_miners)

    def add_curated_miners(
        self, paranet_ual: UAL, miner_addresses: list[Address]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._add_paranet_curated_miners(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            miner_addresses,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _remove_paranet_curated_miners = Method(
        BlockchainRequest.remove_paranet_curated_miners
    )

    def remove_curated_miners(
        self, paranet_ual: UAL, miner_addresses: list[Address]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._remove_paranet_curated_miners(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            miner_addresses,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _request_paranet_curated_miner_access = Method(
        BlockchainRequest.request_paranet_curated_miner_access
    )

    def request_curated_miner_access(
        self, paranet_ual: UAL
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._request_paranet_curated_miner_access(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _approve_curated_miner = Method(BlockchainRequest.approve_curated_miner)

    def approve_curated_miner(
        self, paranet_ual: UAL, miner_address: Address
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._approve_curated_miner(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            miner_address,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _reject_curated_miner = Method(BlockchainRequest.reject_curated_miner)

    def reject_curated_miner(
        self, paranet_ual: UAL, miner_address: Address
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt = self._reject_curated_miner(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            miner_address,
        )

        return {
            "paranetUAL": paranet_ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _get_knowledge_miners = Method(BlockchainRequest.get_knowledge_miners)

    def get_knowledge_miners(
        self, paranet_ual: UAL
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(paranet_ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        paranet_id = Web3.solidity_keccak(
            ["address", "uint256"],
            [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
        )

        return self._get_knowledge_miners(paranet_id)

    _deploy_neuro_incentives_pool = Method(
        BlockchainRequest.deploy_neuro_incentives_pool
    )

    def deploy_incentives_contract(
        self,
        ual: UAL,
        incentives_pool_parameters: NeuroWebIncentivesPoolParams,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> dict[str, str | HexStr | TxReceipt]:
        deploy_incentives_pool_fn = self.incentives_pools_deployment_functions.get(
            incentives_type,
            None,
        )

        if deploy_incentives_pool_fn is None:
            raise ValueError(
                f"{incentives_type} Incentive Type isn't supported. Supported "
                f"Incentive Types: {self.incentives_pools_deployment_functions.keys()}"
            )

        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt: TxReceipt = deploy_incentives_pool_fn(
            knowledge_asset_storage,
            knowledge_asset_token_id,
            **incentives_pool_parameters.to_contract_args(),
        )

        events = self.manager.blockchain_provider.decode_logs_event(
            receipt,
            "ParanetIncentivesPoolFactory",
            "ParanetIncetivesPoolDeployed",
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "incentivesPoolAddress": events[0].args["incentivesPool"]["addr"],
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _get_incentives_pool_address = Method(BlockchainRequest.get_incentives_pool_address)

    def get_incentives_pool_address(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> Address:
        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )
        paranet_id = Web3.solidity_keccak(
            ["address", "uint256"], [knowledge_asset_storage, knowledge_asset_token_id]
        )

        return self._get_incentives_pool_address(paranet_id, incentives_type)

    _register_paranet_service = Method(BlockchainRequest.register_paranet_service)

    def create_service(
        self, ual: UAL, name: str, description: str, addresses: list[Address]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        receipt: TxReceipt = self._register_paranet_service(
            knowledge_asset_storage,
            knowledge_asset_token_id,
            name,
            description,
            addresses,
        )

        return {
            "paranetServiceUAL": ual,
            "paranetServiceId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _add_paranet_services = Method(BlockchainRequest.add_paranet_services)

    def add_services(
        self, ual: UAL, services_uals: list[UAL]
    ) -> dict[str, str | HexStr | TxReceipt]:
        parsed_paranet_ual = parse_ual(ual)
        paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id = (
            parsed_paranet_ual["contract_address"],
            parsed_paranet_ual["token_id"],
        )

        parsed_service_uals = []
        for service_ual in services_uals:
            parsed_service_ual = parse_ual(service_ual)
            (service_knowledge_asset_storage, service_knowledge_asset_token_id) = (
                parsed_service_ual["contract_address"],
                parsed_service_ual["token_id"],
            )

            parsed_service_uals.append(
                {
                    "knowledgeAssetStorageContract": service_knowledge_asset_storage,
                    "tokenId": service_knowledge_asset_token_id,
                }
            )

        receipt: TxReceipt = self._add_paranet_services(
            paranet_knowledge_asset_storage,
            paranet_knowledge_asset_token_id,
            parsed_service_uals,
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [paranet_knowledge_asset_storage, paranet_knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _is_knowledge_miner_registered = Method(
        BlockchainRequest.is_knowledge_miner_registered
    )

    def is_knowledge_miner(self, ual: UAL, address: Address | None = None) -> bool:
        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        paranet_id = Web3.solidity_keccak(
            ["address", "uint256"], [knowledge_asset_storage, knowledge_asset_token_id]
        )

        return self._is_knowledge_miner_registered(
            paranet_id, address or self.manager.blockchain_provider.account.address
        )

    _owner_of = Method(BlockchainRequest.owner_of)

    def is_operator(self, ual: UAL, address: Address | None = None) -> bool:
        knowledge_asset_token_id = parse_ual(ual)["token_id"]

        return self._owner_of(knowledge_asset_token_id) == (
            address or self.manager.blockchain_provider.account.address
        )

    _is_proposal_voter = Method(BlockchainRequest.is_proposal_voter)

    def is_voter(
        self,
        ual: UAL,
        address: Address | None = None,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> bool:
        return self._is_proposal_voter(
            contract=self._get_incentives_pool_contract(ual, incentives_type),
            addr=address or self.manager.blockchain_provider.account.address,
        )

    _get_claimable_knowledge_miner_reward_amount = Method(
        BlockchainRequest.get_claimable_knowledge_miner_reward_amount
    )

    def calculate_claimable_miner_reward_amount(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> int:
        return self._get_claimable_knowledge_miner_reward_amount(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

    _get_claimable_all_knowledge_miners_reward_amount = Method(
        BlockchainRequest.get_claimable_all_knowledge_miners_reward_amount
    )

    def calculate_all_claimable_miner_rewards_amount(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> int:
        return self._get_claimable_all_knowledge_miners_reward_amount(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

    _claim_knowledge_miner_reward = Method(
        BlockchainRequest.claim_knowledge_miner_reward
    )

    def claim_miner_reward(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> dict[str, str | HexStr | TxReceipt]:
        receipt: TxReceipt = self._claim_knowledge_miner_reward(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _get_claimable_paranet_operator_reward_amount = Method(
        BlockchainRequest.get_claimable_paranet_operator_reward_amount
    )

    def calculate_claimable_operator_reward_amount(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> int:
        return self._get_claimable_paranet_operator_reward_amount(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

    _claim_paranet_operator_reward = Method(
        BlockchainRequest.claim_paranet_operator_reward
    )

    def claim_operator_reward(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> dict[str, str | HexStr | TxReceipt]:
        receipt: TxReceipt = self._claim_paranet_operator_reward(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    _get_claimable_proposal_voter_reward_amount = Method(
        BlockchainRequest.get_claimable_proposal_voter_reward_amount
    )

    def calculate_claimable_voter_reward_amount(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> int:
        return self._get_claimable_proposal_voter_reward_amount(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

    _get_claimable_all_proposal_voters_reward_amount = Method(
        BlockchainRequest.get_claimable_all_proposal_voters_reward_amount
    )

    def calculate_all_claimable_voters_reward_amount(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> int:
        return self._get_claimable_all_proposal_voters_reward_amount(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

    _claim_incentivization_proposal_voter_reward = Method(
        BlockchainRequest.claim_incentivization_proposal_voter_reward
    )

    def claim_voter_reward(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> dict[str, str | HexStr | TxReceipt]:
        receipt: TxReceipt = self._claim_incentivization_proposal_voter_reward(
            contract=self._get_incentives_pool_contract(ual, incentives_type)
        )

        parsed_ual = parse_ual(ual)
        knowledge_asset_storage, knowledge_asset_token_id = (
            parsed_ual["contract_address"],
            parsed_ual["token_id"],
        )

        return {
            "paranetUAL": ual,
            "paranetId": Web3.to_hex(
                Web3.solidity_keccak(
                    ["address", "uint256"],
                    [knowledge_asset_storage, knowledge_asset_token_id],
                )
            ),
            "operation": json.loads(Web3.to_json(receipt)),
        }

    def _get_incentives_pool_contract(
        self,
        ual: UAL,
        incentives_type: ParanetIncentivizationType = ParanetIncentivizationType.NEUROWEB,
    ) -> str | dict[str, str]:
        incentives_pool_name = f"Paranet{str(incentives_type)}IncentivesPool"
        is_incentives_pool_cached = (
            incentives_pool_name in self.manager.blockchain_provider.contracts.keys()
        )

        return (
            incentives_pool_name
            if is_incentives_pool_cached
            else {
                "name": incentives_pool_name,
                "address": self.get_incentives_pool_address(ual, incentives_type),
            }
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import itertools
import re
from typing import TYPE_CHECKING, Any, Generic, Type

from dkg.exceptions import ValidationError
from dkg.types import TFunc
from dkg.utils.blockchain_request import (
    ContractInteraction,
    ContractTransaction,
    JSONRPCRequest,
)
from dkg.utils.node_request import NodeCall
from dkg.utils.string_transformations import snake_to_camel

if TYPE_CHECKING:
    from dkg.modules.module import Module
    from dkg.modules.async_module import AsyncModule


class Method(Generic[TFunc]):
    def __init__(self, action: JSONRPCRequest | ContractInteraction | NodeCall):
        self.action = action

    def __get__(
        self,
        obj: "Module | AsyncModule | None" = None,
        _: Type["Module | AsyncModule"] | None = None,
    ) -> TFunc:
        if obj is None:
            raise TypeError(
                "Direct calls to methods are not supported. "
                "Methods must be called from a module instance, "
                "usually attached to a dkg instance."
            )
        return obj.retrieve_caller_fn(self)

    def process_args(self, *args: Any, **kwargs: Any):
        match self.action:
            case JSONRPCRequest():
                return {"args": self._validate_and_map(self.action.args, args, kwargs)}
            case ContractInteraction():
                contract = kwargs.pop("contract", None)
                if not self.action.contract:
                    if contract:
                        self.action.contract = contract
                    else:
                        raise ValidationError(
                            "ContractInteraction requires a 'contract' to be provided"
                        )

                return {
                    "args": self._validate_and_map(self.action.args, args, kwargs),
                    "state_changing": isinstance(self.action, ContractTransaction),
                }
            case NodeCall():
                return self._process_node_call_args(args, kwargs)
            case _:
                return {}

    def _validate_and_map(
        self,
        required_args: dict[str, Type] | Type,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        if not isinstance(required_args, dict):
            if (len(args) + len(kwargs)) != 1:
                raise ValidationError("Exactly one argument must be provided.")

            if len(args) == 1:
                return args[0]
            else:
                return list(kwargs.values())[0]

        if len(args) > len(required_args):
            raise ValidationError(
                "Number of positional arguments can't be bigger than "
                "number of required arguments"
            )

        args_mapped = dict(zip(itertools.islice(required_args.keys(), len(args)), args))
        camel_kwargs = {snake_to_camel(key): value for key, value in kwargs.items()}

        processed_args = {}
        processed_args.update(args_mapped)
        processed_args.update(camel_kwargs)

        if any(missing_params := (arg not in processed_args for arg in required_args)):
            raise ValidationError(
                f"Missing required arg(s): {', '.join(missing_params)}"
            )

        return processed_args

    def _process_node_call_args(
        self, args: list[Any], kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        path_placeholders = re.findall(r"\{([^{}]+)?\}", self.action.path)

        args_in_path = 0
        path_args = []
        path_kwargs = {}
        if len(path_placeholders) > 0:
            for placeholder in path_placeholders:
                if (placeholder != "") and (placeholder in kwargs.keys()):
                    path_kwargs[placeholder] = kwargs.pop(placeholder)
                else:
                    if len(args) <= args_in_path:
                        raise ValidationError(
                            "Number of given arguments can't be smaller than "
                            "number of path placeholders"
                        )

                    if placeholder == "":
                        path_args.append(args[args_in_path])
                    else:
                        path_kwargs[placeholder] = args[args_in_path]

                    args_in_path += 1

        return {
            "path": self.action.path.format(*path_args, **path_kwargs),
            "params": self._validate_and_map(
                self.action.params, args[args_in_path:], kwargs
            )
            if self.action.params
            else {},
            "data": self._validate_and_map(
                self.action.data, args[args_in_path:], kwargs
            )
            if self.action.data
            else {},
        }
